import math
from collections import namedtuple
from pytest import mark, raises
import numpy as np
from statsmodels.stats.descriptivestats import sign_test as sm_sign_test

from combine_pvalues_discrete.tools import (
		is_empty,
		searchsorted_closest,
		has_ties,
		sign_test,
		counted_p, std_from_true_p,
		assert_matching_p_values, assert_discrete_uniform,
	)

@mark.parametrize(
		" argument     , result",[
		( None         , True  ),
		( []           , True  ),
		( ()           , True  ),
		( np.array([]) , True  ),
		( [0]          , False ),
		( [1]          , False ),
		(  0           , False ),
		(  1           , False ),
		( np.array([0]), False ),
		( np.array([1]), False ),
	])
def test_is_empty(argument,result):
	assert is_empty(argument) == result

def test_searchsorted_closest():
	np.testing.assert_array_equal(
		searchsorted_closest( [ 1, 2.5, 3, 7 ], [ 1, 2, 3.1, 9, 0 ] ),
		                                        [ 0, 1,  2 , 3, 0 ]
	)

def test_searchsorted_closest_single_input():
	assert searchsorted_closest([1,2,3],2.1) == 1

@mark.parametrize(
		" argument     , result",[
		( [0,4,1,3,2], False ),
		( [0,1,3,4,1], True  ),
	])
def test_has_ties(argument,result):
	assert has_ties(argument) == result

@mark.parametrize("n",range(1,20))
def test_signtest_with_statsmodels(n):
	# sm: from statsmodels
	X = np.random.normal(size=n)
	
	p = sign_test(X,alternative="two-sided")[0]
	p_sm = sm_sign_test(X)[1]
	
	Y = np.random.normal(n)
	p_2sample = sign_test(X+Y,Y,alternative="two-sided")[0]
	
	assert np.isclose(p,p_sm)
	assert p == p_2sample

@mark.parametrize("n",range(1,20))
def test_signtest_with_statsmodels_onesided(n):
	X = np.random.normal(size=n)
	
	# Because there is no readily available one-sided test:
	if np.mean(X>0)>0.5:
		X = -X
	elif np.mean(X>0)==0.5:
		return

	p = sign_test(X,alternative="less")[0]
	p_sm = sm_sign_test(X)[1]/2

	Y = np.random.normal(n)
	p_2sample = sign_test(X+Y,Y,alternative="less")[0]

	assert np.isclose(p,p_sm)
	assert p == p_2sample

@mark.parametrize("n",range(1,20))
def test_signtest_best_case(n):
	X = np.zeros(n)
	Y = np.ones(n)
	assert np.isclose( sign_test(X,Y,alternative="less")[0], 2**-n )

def test_counted_p():
	null_stats = [1,2,3,4,5,6,7,8,9]
	assert counted_p(0.5,null_stats).pvalue == 0.1
	assert counted_p(3.5,null_stats).pvalue == 0.4
	assert counted_p(10 ,null_stats).pvalue == 1.0
	assert counted_p(0.7,null_stats,atol=0.2).pvalue == 0.1
	assert counted_p(0.9,null_stats,atol=0.2).pvalue == 0.2
	assert counted_p(4.1,null_stats,atol=0.2).pvalue == 0.5
	assert counted_p(4.9,null_stats,atol=0.2).pvalue == 0.6
	assert counted_p(8.6,null_stats,atol=0.2).pvalue == 0.9
	assert counted_p(8.9,null_stats,atol=0.2).pvalue == 1.0

def test_std_counted_p(rng):
	n = 1000  # number of points per dataset
	m = 10000 # number of datasets
	k = 30    # number of different p values tested
	nulls = rng.uniform(0,1,size=(n,m))
	true_ps = np.logspace(-2,0,k)
	
	estimated_ps,estimated_stds = counted_p(
			true_ps[None,None,:],
			nulls[:,:,None],
		)
	
	assert_matching_p_values(
			np.mean(estimated_ps,axis=0),
			true_ps,
			n=n,
			threshold=1e-3,
		)
	assert estimated_ps.shape == (m,k)
	empirical_stds = np.std(estimated_ps,axis=0)
	
	true_stds = std_from_true_p(true_ps,n)
	assert true_stds.shape == (k,)
	
	deviations = true_ps - np.mean(estimated_ps,axis=0)
	# Corrections because the p value is estimated conservatively and, e.g., can never be below 1/(n+1):
	size_offset = (1-true_ps)/(n+1)
	assert np.all( np.abs(deviations+size_offset) <= 3*true_stds/np.sqrt(m) )
	# (√m to get standard error)
	
	np.testing.assert_allclose( true_stds, empirical_stds, rtol=3/np.sqrt(m) )
	np.testing.assert_allclose( true_stds[:-1], np.mean(estimated_stds,axis=0)[:-1], rtol=3/np.sqrt(m) )
	# Last element is expected to be unequal, because the estimate cannot reasonably be zero.

def test_assert_matching_p_values_one_sample():
	assert_matching_p_values(0.04,0.05,n=30)
	assert_matching_p_values([0.04,0.051,0.048],0.05,n=30)
	assert_matching_p_values([0.04,0.061,0.02],[0.05,0.06,0.02],n=30)
	with raises(AssertionError):
		assert_matching_p_values([0.047,0.061,0.02],[0.05,0.01,0.02],n=300)

def test_assert_matching_p_values_two_sample():
	assert_matching_p_values(0.04,0.05,n=30,compare=True)
	assert_matching_p_values([0.04,0.051,0.048],0.05,n=30,compare=True)
	assert_matching_p_values([0.04,0.061,0.02],[0.05,0.06,0.02],n=(30,40),compare=True)
	with raises(AssertionError):
		assert_matching_p_values([0.047,0.061,0.02],[0.05,0.01,0.02],n=3000,compare=True)
	with raises(AssertionError):
		assert_matching_p_values(0.1,0.9,n=10,compare=True)



@mark.parametrize("size",range(10,100,10))
@mark.parametrize("n_values",range(2,10))
def test_assert_discrete_uniform(size,n_values,rng):
	moderate_value = rng.uniform(0.3,0.7)
	values = sorted( np.hstack(( rng.random(n_values-2), [moderate_value,1] )) )
	probs = np.diff(values,prepend=0)
	data = rng.choice( values, p=probs, size=size, replace=True )
	assert_discrete_uniform(data,threshold=1e-4)

def test_assert_discrete_uniform_perfect():
	data = [ 0.1, 0.3, 0.3, 0.7, 0.7, 0.7, 0.7, 1.0, 1.0, 1.0 ]
	assert_discrete_uniform(data,threshold=0.5)

def test_assert_discrete_uniform_fail():
	data = [ 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.7, 0.7, 0.7 ]
	with raises(AssertionError):
		assert_discrete_uniform(data)




#!/usr/bin/python3
# -*- coding: utf-8 -*-

"""
In this example, we illustrate the benefits of this module by comparing it with other approaches to the same dataset, all of which yield wrong results or consume a lot of time.

Suppose, we want to to explore the effect on a drug on dogs.
We expect our observable (by which we measure the effect of the drug) be more affected by the dog breed than by the drug, e.g., because a poodle is generally weaker than a mastiff.
Therefore we group the dogs by breed a priori, creating sub-datasets.
Each breed group gets further randomly split into a treatment (T) and control group (C).
Since not all dogs complete the study, our sub-datasets become very inhomogeneous in sample size.

Our data looks like this:

.. literalinclude:: ../examples/comparison.py
	:start-after: example-st\u0061rt
	:dedent: 1
	:lines: 1-12

Each pair represents one breed, with the first half being the control group and the second the treatment group.
If our drug works as desired, the second half should exhibit higher values.
Finally let’s assume that due to the nature of our observable, we only want to use a ranked statistics.

Thus, we want to investigate the null hypothesis that for each sub-dataset, both samples are from the same distribution (or more precisely, they fulfil the null hypothesis of the Mann–Whitney *U* test).
The alternative hypothesis is that that the first pair of samples are from a distribution with a lower median.

First, suppose we discard our information on breeds and pool the control and treatment groups.
We then apply the Mann–Whitney *U* test to the pooled samples.
This way, we do not need to combine tests, but we lose statistical power, i.e., we increase the false-negative rate.

.. literalinclude:: ../examples/comparison.py
	:start-after: example-st\u0061rt
	:dedent: 1
	:lines: 15-20

Alternatively, we can summarise the samples in each pair by their median and use the sign test to compare the groups.
Again, we discard information and thus lose statistical power:

.. literalinclude:: ../examples/comparison.py
	:start-after: example-st\u0061rt
	:dedent: 1
	:lines: 22-29

To properly take into account all information, we have to apply the Mann–Whitney *U* test to each pair (breed) and then combine the *p* values.
SciPy’s `combine_pvalues` allows us to do this, but it requires continuous tests.
Since the Mann–Whitney *U* test does not fulfil this requirement, we will overestimate the combined *p* value:

.. literalinclude:: ../examples/comparison.py
	:start-after: example-st\u0061rt
	:dedent: 1
	:lines: 32-37

Finally, by using this module, we can take into account the discreteness of tests, obtaining a correct combined *p* value:

.. literalinclude:: ../examples/comparison.py
	:start-after: example-st\u0061rt
	:dedent: 1
	:lines: 40-44

We here use Fisher’s method, since it simplifies the following demonstration.
See `method_choice` as to how this affects our results and their interpretation.

Checking the result
```````````````````

Let’s convince ourselves that the result of `combine` is actually correct.
To this end, we first implement the statistic of Fisher’s method (as used by SciPy) for combining Mann–Whitney *U* tests.
Note how the result agrees with that of applying `combine_pvalues` above:

.. literalinclude:: ../examples/comparison.py
	:start-after: example-st\u0061rt
	:dedent: 1
	:lines: 47-53


Next, we implement a function that samples analogous datasets corresponding to our null hypothesis (surrogates).
(Since we only care about the order of samples, we do not have to recreate the magnitude of values.)

.. literalinclude:: ../examples/comparison.py
	:start-after: example-st\u0061rt
	:dedent: 1
	:lines: 55-60

Finally, we sample `n=10000` times from our null model and estimate the *p* value by comparing the values of Fisher’s statistic for the null model and the original data.

.. literalinclude:: ../examples/comparison.py
	:start-after: example-st\u0061rt
	:dedent: 1
	:lines: 62-67

This confirms the low *p* value we obtained with `combine` above and that the *p* values obtained with the other methods were too high.
You may note that, although this value is low, it is not within the confidence interval of the result of `combine` from above.
The reason for this is that `n` and thus `count` is low (in the above, `count` is just 16).
Thus, the *p* value estimate from the null model is subject to much higher fluctuations than that of `combine`.
Obtaining a precision comparable to `combine` would require an excessive amount of time.

Once more with weights
``````````````````````

So far, we have treated the *p* values for each sub-dataset equally.
However, the different sub-datasets differ considerably in size and thus meaningfulness, e.g., the first contains seven data points while the last only contains two.
We can account for this by weighting the sub-datasets with their degrees of freedom (number of samples minus one), which are stored with the test results for this purpose:

.. literalinclude:: ../examples/comparison.py
	:start-after: example-st\u0061rt
	:dedent: 1
	:lines: 69-70

And like above, we can check this result by comparing to an explicit, computationally expensive simulation of the null hypothesis:

.. literalinclude:: ../examples/comparison.py
	:start-after: example-st\u0061rt
	:dedent: 1
	:lines: 72-85

"""

if __name__ == "__main__":
	# example-start
	data = [# control group  , treatment group
	        ( [20,44,14,68]  , [73,22,80]    ), # Breed 1
	        ( [60,68,46,45]  , [30]          ), # Breed 2
	        ( [92,97,98]     , [84,89]       ), # …
	        ( [14]           , [21,45,31,23] ),
	        ( [24,58,0,24,33], [65,51,61]    ),
	        ( [93,76,70,83]  , [84]          ),
	        ( [10,2]         , [28,36,11]    ),
	        ( [27]           , [38,58]       ),
	        ( [8,13,37]      , [43,51]       ),
	        ( [18]           , [12]          ),
	]
	
	# Pooling data and MWU test
	from scipy.stats import mannwhitneyu
	
	pooled_Cs = [ c for C,T in data for c in C ]
	pooled_Ts = [ t for C,T in data for t in T ]
	print( mannwhitneyu(pooled_Cs,pooled_Ts,alternative="less") )
	# MannwhitneyuResult(statistic=282.0, pvalue=0.30908071682819527)
	
	# Summarizing data and sign test
	import numpy as np
	from combine_pvalues_discrete import sign_test
	
	reduced_Cs = [ np.median(C) for C,T in data ]
	reduced_Ts = [ np.median(T) for C,T in data ]
	print( sign_test( reduced_Cs, reduced_Ts, alternative="less" ) )
	# SignTestResult(pvalue=0.171875, not_tied=10, statistic=3)
	
	# Combining MWU results without respecting discreteness
	from scipy.stats import combine_pvalues, mannwhitneyu
	
	pvalues = [ mannwhitneyu(C,T,alternative="less").pvalue for C,T in data ]
	statistic,pvalue = combine_pvalues(pvalues,method="fisher")
	print(statistic,pvalue)
	# (27.447712265267114, 0.123131292229715)
	
	# Combining MWU results with respecting discreteness
	from combine_pvalues_discrete import CTR, combine
	
	ctrs = [ CTR.mann_whitney_u(C,T,alternative="less") for C,T in data ]
	print( combine(ctrs,method="fisher") )
	# Combined_P_Value(pvalue=0.0014229998577000142, std=1.1920046440408576e-05)
	
	# Demonstrating correctness with null-model simulation
	def fisher_statistic(dataset):
		pvalues = [ mannwhitneyu(C,T,alternative="less").pvalue for C,T in dataset ]
		return -2*np.sum(np.log(pvalues))
	
	data_statistic = fisher_statistic(data)
	print(data_statistic)
	# 27.447712265267114
	
	rng = np.random.default_rng()
	def null_sample(data):
		return [
			( rng.random(len(C)), rng.random(len(T)) )
			for C,T in data
		]
	
	n = 10000
	null_statistic = [ fisher_statistic(null_sample(data)) for _ in range(n) ]
	count = np.sum( null_statistic >= data_statistic )
	p_from_simulation = (count+1)/(n+1)
	print(p_from_simulation)
	# 0.0016998300169983002
	
	print( combine(ctrs,method="fisher",weights="dof") )
	# Combined_P_Value(pvalue=0.0011117998888200112, std=1.0537855099673787e-05)
	
	def weighted_fisher_statistic(dataset,weights):
		pvalues = [ mannwhitneyu(C,T,alternative="less").pvalue for C,T in dataset ]
		return -2*weights.dot(np.log(pvalues))
	
	weights = np.array([len(C)+len(T)-1 for C,T in data])
	weighted_data_statistic = weighted_fisher_statistic(data,weights)
	weighted_null_statistic = [
			weighted_fisher_statistic(null_sample(data),weights)
			for _ in range(n)
		]
	weighted_count = np.sum( weighted_null_statistic >= weighted_data_statistic )
	weighted_p_from_simulation = (weighted_count+1)/(n+1)
	print(weighted_p_from_simulation)
	# 0.0010998900109989002
	



import sys
import os
from setuptools_scm import get_version
from unittest.mock import MagicMock as Mock

# Mocking to make RTD autobuild the documentation.
autodoc_mock_imports = ["numpy"]
#sys.modules.update([("numpy", Mock())])
sys.path.insert(0,os.path.abspath("../combine_pvalues_discrete"))
sys.path.insert(0,os.path.abspath("../examples"))

needs_sphinx = "1.6"

extensions = [
	"sphinx.ext.autodoc",
	"sphinx.ext.autosummary",
	"sphinx.ext.mathjax",
	"matplotlib.sphinxext.plot_directive",
	"numpydoc",
]

source_suffix = ".rst"

master_doc = "index"

project = "Combine p Values Discrete"
copyright = "Gerrit Ansmann"

release = version = get_version(root="..", relative_to=__file__)

default_role = "any"

add_function_parentheses = True

add_module_names = False

html_theme = "nature"
pygments_style = "colorful"
htmlhelp_basename = "Combine_p_Values_Discrete_doc"

numpydoc_show_class_members = False
autodoc_member_order = "bysource"

toc_object_entries_show_parents = 'hide'

def on_missing_reference(app, env, node, contnode):
	if node["reftype"] == "any":
		return contnode
	else:
		return None

def setup(app):
	app.connect("missing-reference", on_missing_reference)




from collections import namedtuple
import numpy as np
from scipy.stats import binomtest, kstwo, boschloo_exact, fisher_exact

def is_empty(x):
	"""
	Whether the argument is an empty 1D iterable or None.
	"""
	try:
		return len(x)==0
	except TypeError:
		return x is None

def searchsorted_closest(array,values):
	"""
	Wrapper around NumPy’s `searchsorted` that returns the index of the closest value(s) – as opposed to the next lower or higher one.
	"""
	array = np.asarray(array)
	interval = (0,len(array)-1)
	right_idcs = np.searchsorted(array,values,side="left").clip(*interval)
	left_idcs = (right_idcs-1).clip(*interval)
	
	left_or_right = values-array[left_idcs] < array[right_idcs]-values
	return np.choose( left_or_right, (right_idcs,left_idcs) )

def has_ties(array):
	"""
	Whether any two values in the array are identical (tied).
	"""
	return np.any(np.diff(sorted(array))==0)

SignTestResult = namedtuple("SignTestResult",("pvalue","not_tied","statistic"))

def sign_test(x,y=0,alternative="less"):
	"""
	Just the sign test without any combination features, provided because I have it anyway.
	
	**two-sided:**
	Pass paired samples `x` and `y` as arguments. The tested null hypothesis is that `x[i]` and `y[i]` are from the same distribution (separately for each `i`).
	
	**one-sided**
	Pass a single sample `x` and a number `y`. The tested null hypothesis is that `x` is sampled from a distribution with a median larger than `y`.
	
	Returns a named tuple consisting of the p value, the number of non-tied samples, and the statistic of the sign test (number of samples which are greater than the reference).
	
	Examples
	--------
	.. code-block:: python3
	
		dataset_A = [20,44,14,68]
		dataset_B = [73,22,80,53]
		p = sign_test(dataset_A,dataset_B,alternative="less").pvalue
	"""
	
	x = np.asarray(x)
	y = np.asarray(y)
	greater = np.sum(x>y)
	less    = np.sum(x<y)
	non_tied = less+greater
	return SignTestResult(
			binomtest( greater, non_tied, alternative=alternative ).pvalue,
			non_tied,
			greater,
		)

def count_greater_or_close(x,y,atol=0,rtol=0):
	"Counts how often x is greater than y or close with atol or rtol."
	
	comparison = (y<=x)
	
	if atol or rtol:
		comparison |= np.isclose(x,y,atol=atol,rtol=rtol)
	
	return np.sum(comparison,axis=0)

Combined_P_Value = namedtuple("Combined_P_Value",("pvalue","std"))

def counted_p(orig_stat,null_stats,**tols):
	"""
	Estimates the p value of a statistic (`orig_stat`) by comparing with the statistic for samples of a null model (`null_stats`), with a small statistic being extreme. Returns the p value and its (estimated) standard deviation when estimating with this method.
	"""
	
	null_stats = np.asarray(null_stats)
	size = null_stats.shape[0]
	count = count_greater_or_close(orig_stat,null_stats,**tols)
		
	p = (count+1)/(size+1)
	std = np.maximum(
			np.sqrt(count*(1-count/size))/(size+1),
			1/(size+1),
		)
	return Combined_P_Value(p,std)

def std_from_true_p(true_p,size):
	"""
	Standard deviation of p value from samples, if the true p value is known.
	"""
	return np.sqrt(true_p*(1-true_p)*size)/(size+1)

def assert_matching_p_values(tested_ps,target_ps,n,threshold=1e-4,compare=False):
	"""
	Asserts that `ps` (estimated with `counted_p`) matches `target_ps` when estimated from `n` samples of the null model using the binomial test.
	
	`threshold` is the maximal *p* value of that binomial test.
	
	If `target_ps` is not exact but estimated by sampling as well (with the same `n`), set `compare=True`. In this case, Boschloo’s or Fisher’s exact test is used for comparison, with the latter being used for n>10 because it’s considerably faster, though less accurate. If the sample sizes are different, `n` needs to be a pair.
	"""
	
	tested_ps = np.atleast_1d(tested_ps)
	assert tested_ps.ndim == 1
	assert np.all(tested_ps<=1)
	assert np.all(0<=tested_ps)
	
	if np.ndim(target_ps)==0:
		target_ps = np.full_like(tested_ps,target_ps)
	
	if compare:
		for i,ps in enumerate(zip(tested_ps,target_ps)):
			ns = [n,n] if np.ndim(n) == 0 else n
			
			# If you wonder why we are reconstructing the ks here: It’s because we are mostly testing functions that return p values, not ks and ns.
			ks = [ round(p*(n+1)-1) for p,n in zip(ps,ns) ]
			comparison_test = boschloo_exact if max(ns)<=10 else fisher_exact
			comparison_p = comparison_test([
					ks,
					[n-k for k,n in zip(ks,ns)],
				]).pvalue
			if comparison_p < threshold:
				ratios = [f"{p} = ({k}+1)/({n}+1)" for p,n,k in zip(ps,ns,ks)]
				raise AssertionError(
					f"""
					p values don’t match.
						ratios[0] ≉ ratios[1]
						p value of {comparison_test.__name__}: {comparison_p}
					""")
	else:
		for i,(tested_p,target_p) in enumerate(zip(tested_ps,target_ps)):
			assert np.ndim(n) == 0
			k = round(tested_p*(n+1)-1)
			binomtest_p = binomtest(k,n,target_p).pvalue
			if binomtest_p < threshold:
				raise AssertionError(
					f"""
					p values don’t match.
						{tested_p} = ({k}+1)/({n}+1) ≉ {target_p}
						p value of binomial test: {binomtest_p}
					""")

def assert_discrete_uniform(data,threshold=1e-4):
	"""
	Checks whether `data` complies with a discrete uniform distribution on the interval [0,1]. `threshold` is the maximum *p* value of the binomial test used under the hood.
	"""
	data = np.asarray(data)
	n = len(data)
	values = set(data)
	if len(values)<2:
		raise ValueError("Need at least two distinct values.")
	
	for value in values:
		assert_matching_p_values(
				count_greater_or_close(value,data,atol=1e-15,rtol=1e-15)/n,
				value,
				n = n,
				threshold = threshold,
				compare = False
			)




from pytest import mark, raises
from itertools import count, product
import numpy as np
from scipy.stats import (
		mannwhitneyu,
		wilcoxon,
		spearmanr, kendalltau,
		fisher_exact, boschloo_exact,
	)
import math

from combine_pvalues_discrete.ctr import CTR, combine, combining_statistics
from combine_pvalues_discrete.tools import (
		sign_test,
		counted_p,
		assert_matching_p_values, assert_discrete_uniform,
	)

n_samples = 100000

def test_simple_mwu():
	assert (
		CTR.mann_whitney_u([0],[1],alternative="less")
		==
		CTR( 0.5, [0.5,1.0], 1 )
	)

def test_simple_signtest():
	assert (
		CTR.sign_test([0],[1],alternative="less")
		==
		CTR( 0.5, [0.5,1.0], 1 )
	)

@mark.parametrize("n",range(3,10))
def test_simple_wilcoxon(n,rng):
	data = np.arange(1,n+1)*rng.choice([-1,1],n)
	
	wilcoxon_kwargs = dict(alternative="less",mode="exact")
	control_p = wilcoxon(data,None,**wilcoxon_kwargs).pvalue
	control_all_ps = list({
			wilcoxon( np.arange(1,n+1)*signs, **wilcoxon_kwargs ).pvalue
			for signs in product([-1,1],repeat=n)
		})
	
	assert (
		CTR.wilcoxon_signed_rank(data,alternative="less")
		==
		CTR( control_p, control_all_ps, n )
	)

@mark.parametrize(
		"  x    ,    y   ,    alt   ,   p  ,      all_ps         ",
	[
		([1,3,2], [4,5,0], "less"   ,  5/6 , [ 1/6, 1/2, 5/6, 1 ]),
		([1,3,2], [4,5,0], "greater",  1/2 , [ 1/6, 1/2, 5/6, 1 ]),
		([1,2,2], [3,3,5], "less"   ,   1  , [ 1/3,           1 ]),
		([1,2,2], [3,3,5], "greater",  2/3 , [ 2/3,           1 ]),
	])
def test_simple_spearman(x,y,alt,p,all_ps):
	result = CTR.spearmanr( x, y, alternative=alt )
	control = CTR(p,all_ps,len(x)-1)
	assert result._approx(control)

@mark.parametrize("alt",["less","greater"])
def test_spearman_large_dataset(rng,alt):
	n = 100
	x,y = rng.normal(size=(2,n))
	ctr = CTR.spearmanr(x,y,alternative=alt)
	assert ctr.p == spearmanr(x,y,alternative=alt).pvalue
	assert ctr.nulldist.continuous
	assert ctr.dof == n-1

def test_spearman_large_perfect_dataset():
	n = 100
	ctr = CTR.spearmanr(range(n),range(n),alternative="greater")
	assert ctr.p >= 1/math.factorial(n)

@mark.parametrize("n",range(2,9))
def test_spearman_nulldist_length(n):
	assert (
		len( CTR.spearmanr(range(n),range(n)).nulldist.ps )
		==
		math.comb(n+1,3) + (n!=3) # OEIS A126972
	)

def spearman_data(RNG,n,trend=0):
	x,y = RNG.normal(size=(2,n))
	y = (1-trend)*y + trend*x
	return x,y

@mark.parametrize("alt",["less","greater"])
@mark.parametrize("n",range(3,7))
def test_spearman_null(n,alt,rng):
	m = 1000 if n<5 else 100
	p_values = [
			CTR.spearmanr(
				*spearman_data(RNG=rng,n=n),
				alternative = alt
			).p
			for _ in range(m)
		]
	
	assert_discrete_uniform(p_values)

@mark.parametrize("n",range(3,9))
def test_spearman(n,rng):
	m = 1000
	
	x,y = spearman_data(RNG=rng,n=n,trend=0.8)
	orig_ρ = spearmanr(x,y).correlation
	
	null_ρs = np.array([
			spearmanr(*spearman_data(RNG=rng,n=n)).correlation
			for _ in range(m)
		])
	
	assert_matching_p_values(
			counted_p( orig_ρ, null_ρs ).pvalue,
			CTR.spearmanr(x,y,alternative="less").p,
			n = m,
		)

@mark.parametrize(
		"    x    ,     y    ,    alt   ,    p   ,                all_ps                 ",
	[
		([1,3,2,4], [4,5,0,6], "less"   ,  23/24 , [ 1/24, 1/6, 3/8, 5/8, 5/6, 23/24, 1 ]),
		([1,3,2,4], [4,5,0,6], "greater",   1/6  , [ 1/24, 1/6, 3/8, 5/8, 5/6, 23/24, 1 ]),
	])
def test_simple_kendall(x,y,alt,p,all_ps):
	result = CTR.kendalltau(x,y,alternative=alt)
	control_p = kendalltau(x,y,alternative=alt).pvalue
	n = len(x)
	assert result._approx( CTR(p,all_ps,n-1) )
	assert np.isclose( result.p, control_p )

@mark.parametrize(
		"      C      ,   alt    ,    p    ,            all_ps            ",
	[
		([[2,3],[4,0]], "less"   ,   5/42  , [  5/42 , 25/42 ,  20/21 , 1 ]),
		([[2,3],[4,0]], "greater",    1    , [  1/21 , 17/42 ,  37/42 , 1 ]),
		([[1,7],[2,7]], "less"   ,  93/170 , [ 21/170, 93/170,  78/85 , 1 ]),
		([[1,7],[2,7]], "greater", 149/170 , [  7/85 , 77/170, 149/170, 1 ]),
	])
def test_simple_fisher_exact(C,alt,p,all_ps):
	result = CTR.fisher_exact( C, alternative=alt )
	control_p = fisher_exact(C,alternative=alt)[1]
	assert result._approx( CTR(p,all_ps,np.sum(C)) )
	assert np.isclose( result.p, control_p )

def test_simple_boschloo():
	C = [[1,2],[3,0]]
	result = CTR.boschloo_exact( C, alternative="less" )
	control_p = boschloo_exact(C,alternative="less" ).pvalue
	assert np.isclose( result.p, control_p )
	all_ps = [ 1/64, 0.079305, 0.273032, 11/32, 0.57860, 49/64, 1 ]
	control = CTR(0.079305,all_ps,np.sum(C))
	assert result._approx(control,atol=1e-5)

# -----------------

# What follows are extensive simulations checking the null distributions of the combined p values as well as comparing the combined p values with those obtained by extensive null models. We first prepare utility functions for two tests (sign test and MWU test). We use Fisher’s method, hence the sums of logarithms of p values (logp_sum).

def mwu_combine( data, **kwargs ):
	ctrs = [ CTR.mann_whitney_u(X,Y,alternative="less") for X,Y in data ]
	return combine(ctrs,**kwargs).pvalue

def mwu_data(RNG,n,trend=0):
	"""
		Creates a dataset of `n` unequal pairs of normally distributed numbers with a trend towards the first half of a pair containing smaller values.
		If `trend` is zero, this conforms with the null hypothesis.
	"""
	return [(
			RNG.normal(size=RNG.randint(2,6))-trend,
			RNG.normal(size=RNG.randint(2,6))
		) for _ in range(n) ]

def mwu_logp_sum(data):
	return sum(
		np.log10(mannwhitneyu(X,Y,alternative="less",method="exact").pvalue)
		for X,Y in data
	)

def mwu_invert(data):
	return [ (-X,-Y) for X,Y in data ]

def signtest_combine( data, **kwargs ):
	ctrs = [ CTR.sign_test(X,Y,alternative="less") for X,Y in data ]
	return combine(ctrs,**kwargs).pvalue

def signtest_data(RNG,n,trend=0):
	"""
		Creates a dataset of `n` pairs of normally distributed numbers with a trend towards the first half of a pair containing smaller values.
		If `trend` is zero, this conforms with the null hypothesis.
	"""
	return [(
			RNG.normal(size=size)-trend,
			RNG.normal(size=size)
		) for size in RNG.randint(15,21,size=n) ]

def signtest_logp_sum(data):
	return sum(
		np.log10(sign_test(X,Y,alternative="less")[0])
		for X,Y in data
	)

def signtest_invert(data):
	return [ (Y,X) for X,Y in data ]

tests = {
		"signtest": ( signtest_combine, signtest_data, signtest_logp_sum, signtest_invert ),
		"mwu_test": (      mwu_combine,      mwu_data,      mwu_logp_sum,      mwu_invert ),
	}

@mark.slow
@mark.parametrize("method,variant",combining_statistics)
@mark.parametrize("sampling_method",["proportional","stochastic"])
@mark.parametrize("test",tests)
@mark.parametrize("alt",["less","greater","two-sided"])
def test_null_distribution(method,variant,test,sampling_method,alt,rng):
	test_and_combine,create_data,*_ = tests[test]
	n = 10
	p_values = [
		test_and_combine(
			create_data(rng,n),
			alternative = alt,
			RNG = rng,
			method = method,
			n_samples = 1000,
			sampling_method = sampling_method,
			weights = rng.random(n) if variant=="weighted" else None
		)
		for _ in range(50)
	]
	
	assert_discrete_uniform(p_values)

def create_surrogate(RNG,pairs):
	"""
	Creates a single artificial dataset complying with the null hypothesis (surrogate).
	This dataset has the same shape as `pairs`.
	"""
	return [
		[ RNG.normal(size=len(member)) for member in pair ]
		for pair in pairs
	]

@mark.parametrize("trend",np.linspace(-0.7,0.7,10))
@mark.parametrize("sampling_method",["proportional","stochastic"])
@mark.parametrize("test",tests)
@mark.parametrize("alt",["less","greater","two-sided"])
def test_compare_with_surrogates(trend,test,sampling_method,alt,rng):
	test_and_combine,create_data,logp_sum,invert = tests[test]
	dataset = create_data(rng,10,trend=trend)
	
	p_from_combine = test_and_combine(
			dataset,
			method = "fisher",
			n_samples = n_samples,
			alternative = alt,
			RNG = rng
		)
	
	n = 100
	
	evaluate = {
		"less":      lambda data: logp_sum(data),
		"greater":   lambda data: logp_sum(invert(data)),
		"two-sided": lambda data: min( logp_sum(data), logp_sum(invert(data)) ),
	}[alt]
	
	original_logp_sum = evaluate(dataset)
	surrogate_logp_sums = [
		evaluate( create_surrogate(rng,dataset) )
		for _ in range(n)
	]
	p_from_surrogates = counted_p( original_logp_sum, surrogate_logp_sums ).pvalue
	
	assert_matching_p_values(
			p_from_combine,
			p_from_surrogates,
			n = (n_samples,n),
			compare=True,
		)




from pytest import mark, raises
import numpy as np
from scipy.stats import uniform, ks_1samp, binomtest
from itertools import combinations

from combine_pvalues_discrete.pdist import PDist, sample_discrete, sample_uniform
from combine_pvalues_discrete.tools import assert_matching_p_values

@mark.parametrize("method",("stochastic","proportional"))
def test_sample_discrete(method,rng):
	threshold = 0.5 if method=="proportional" else 1e-3
	m = 10
	n = 100000
	values = range(m)
	frequencies = rng.uniform(0,1,size=m)
	frequencies /= np.sum(frequencies)
	
	result = sample_discrete(values,frequencies,RNG=rng,size=n)
	for value,frequency in zip(values,frequencies):
		count = np.sum(result==value)
		assert binomtest(count,n,frequency).pvalue>threshold

def test_core_stuff():
	dists = ( PDist([]), PDist([1]), PDist([0.5,1]) )
	for i,dist in enumerate(dists):
		if i==0:
			assert dist.continuous
		assert dist == dist
	for dist1,dist2 in combinations(dists,2):
		assert dist1 != dist2

def test_no_end_at_1():
	with raises(ValueError):
		PDist([0.1,0.5])

def test_negative_p():
	with raises(ValueError):
		PDist([-0.1,0.5])

def test_zero_p():
	with raises(ValueError):
		PDist([0.1,0.5,0])

def test_p_more_than_one():
	with raises(ValueError):
		PDist([-0.1,0.5,2])

def test_correct_almost_1():
	dist = PDist([0.1,1+1e-14])
	assert dist.ps[-1] == 1

def test_p_double_one():
	with raises(ValueError):
		PDist([0.1,0.5,1,1+1e-14])


@mark.parametrize("size",range(1,100))
def test_normalisation(size):
	dist = PDist( list(np.random.random(size-1)) + [1] )
	assert np.sum(dist.probs) == 1

@mark.parametrize("size",range(1,100))
def test_cumprobs(size):
	dist = PDist( list(np.random.random(size-1)) + [1] )
	np.testing.assert_almost_equal( dist.ps, np.cumsum(dist.probs) )

@mark.parametrize("n_ps",2**np.arange(0,10))
@mark.parametrize("n_samples",10**np.arange(4,7))
@mark.parametrize("method",("stochastic","proportional"))
def test_sampling(n_ps,n_samples,method,rng):
	threshold = 0.5 if method=="proportional" else 1e-4
	if n_ps>1:
		dist = PDist( list(rng.random(n_ps-1)) + [1] )
		sample = dist.sample(RNG=rng,size=n_samples,method=method)
		for p,prob in zip(dist,dist.probs):
			assert np.isclose( np.average(sample==p), prob, atol=3/np.sqrt(n_samples) )
			assert binomtest(np.sum(sample<=p),n_samples,p).pvalue > threshold
	else:
		dist = PDist([])
		sample = dist.sample(RNG=rng,size=n_samples,method=method)
		assert ks_1samp(sample,uniform.cdf).pvalue > threshold

@mark.parametrize("n_ps",2**np.arange(0,10))
@mark.parametrize("n_samples",10**np.arange(4,7))
@mark.parametrize("method",("stochastic","proportional"))
def test_sampling_complement(n_ps,n_samples,method,rng):
	threshold = 0.5 if method=="proportional" else 1e-4
	if n_ps>1:
		dist = PDist( list(rng.random(n_ps-1)) + [1] )
		sample = dist.sample_complement(RNG=rng,size=n_samples,method=method)
		for p,prob in zip(dist,dist.probs):
			q = dist.complement(p)
			assert np.isclose( np.average(sample==q), prob, atol=3/np.sqrt(n_samples) )
			assert binomtest(np.sum(sample<=q),n_samples,q).pvalue >= threshold
	else:
		dist = PDist([])
		sample = dist.sample_complement(RNG=rng,size=n_samples,method=method)
		assert ks_1samp(sample,uniform.cdf).pvalue > threshold

@mark.parametrize("n_ps",2**np.arange(0,10))
@mark.parametrize("method",("stochastic","proportional"))
def test_sampling_both(n_ps,method,rng):
	if n_ps>1:
		dist = PDist( list(rng.random(n_ps-1)) + [1] )
	else:
		dist = PDist([])
	
	sampling_methods = {
			"p": dist.sample,
			"q": dist.sample_complement,
			"b": dist.sample_both,
		}
	samples = {}
	for key,sampling_method in sampling_methods.items():
		RNG = np.random.default_rng(42)
		samples[key] = sampling_method(RNG=RNG,size=10000,method=method)
	assert np.all( samples["p"] == samples["b"][0] )
	assert np.all( samples["q"] == samples["b"][1] )

def test_complement():
	dist = PDist([0.1,0.5,1])
	assert dist.complement(0.1)==1
	assert dist.complement(0.5)==0.9
	assert dist.complement(1)==0.5




#!/usr/bin/python3
# -*- coding: utf-8 -*-

"""
Suppose, we want to implement the Mann–Whitney *U* test (with the alternative “less”) into the CTR framework.
(This test is already implemented, so we don’t actually need to do this.)

The test is named for its statistics *U* which can exhibit values between 0 and :math:`n·m`, where :math:`n` and :math:`m` are the sizes of the datasets that are compared.
These *U* values are then translated into *p* values.
If we look into SciPy’s implementation of this test, we find that it uses an implementation of the test’s null distribution (`scipy.stats._mannwhitneyu._mwu_state`), which we can exploit.
Finally, we can use SciPy’s implementation to compute the *p* value of our given datasets.

With that we have all the ingredients to write a small function to return a combinable test result for a given pair of datasets:

.. literalinclude:: ../examples/mwu.py
	:start-after: example-st\u0061rt
	:dedent: 1
	:lines: 1-8
"""

if __name__ == "__main__":
	import numpy as np
	
	data_A = np.random.random(7)
	data_B = np.random.random(8)
	# example-start
	from combine_pvalues_discrete import CTR
	from scipy.stats._mannwhitneyu import _mwu_state, mannwhitneyu
	
	def mwu_ctr(x,y):
		p = mannwhitneyu(x,y,method="exact",alternative="less").pvalue
		n,m = len(x),len(y)
		possible_ps = [ _mwu_state.cdf(U,n,m) for U in range(n*m+1) ]
		return CTR( p, possible_ps )
	
	print(mwu_ctr(data_A,data_B))
	



from pytest import mark
import numpy as np

from combine_pvalues_discrete.ctr import CTR, direction, combining_statistics

@mark.parametrize(
			"  ps  , method, result",
		[
			([0.1,0.1,0.9], "fisher" , "less"   ),
			([0.1,0.9,0.9], "fisher" , "greater"),
			([0.1,0.5,0.9], "fisher" , "equal"  ),
			([0.1,0.99],    "fisher" , "greater"),
			([0.1,0.99],    "pearson", "greater"),
		]
	)
def test_simple(ps,method,result):
	ctrs = [ CTR(p) for p in ps ]
	assert direction(ctrs,method=method) == result

@mark.parametrize("method,variant",combining_statistics)
def test_big(method,variant,rng):
	n = 100
	ctrs = [ CTR.sign_test(rng.normal(-1,1,10)) for _ in range(n) ]
	weights = rng.random(n) if variant=="weighted" else None
	assert direction(ctrs,weights,method) == "less"




from pytest import mark, raises
from itertools import chain, product, combinations_with_replacement
import numpy as np
from math import prod, sqrt

from scipy.stats import combine_pvalues
from scipy.special import erf, erfinv

from combine_pvalues_discrete.ctr import CTR, combine, combining_statistics, assert_one_sided
from combine_pvalues_discrete.tools import sign_test, assert_matching_p_values

n_samples = 10000

examples = [
	CTR( 0.5, [0.5,      1] ),
	CTR( 1.0, [0.5,      1] ),
	CTR( 0.3, [0.3, 0.5, 1] ),
	CTR( 0.7, [0.2, 0.7, 1] ),
]

@mark.parametrize(
	  "  p  , alternative, result",
	[
	  ( 0.01, "less"     , 0.01   ),
	  ( 0.01, "greater"  , 0.99   ),
	  ( 0.01, "two-sided", 0.02   ),
	  ( 0.4 , "two-sided", 0.8    ),
	  ( 0.99, "less"     , 0.99   ),
	  ( 0.99, "greater"  , 0.01   ),
	  ( 0.99, "two-sided", 0.02   ),
	  ( 0.6 , "two-sided", 0.8    ),
	])
def test_single_p(p,alternative,result):
	assert np.isclose(
			combine([CTR(p)],alternative=alternative).pvalue,
			result
		)

def test_single_p_extreme():
	assert combine([examples[3]],alternative="two-sided").pvalue == 1

def test_assert_one_sided_good():
	assert_one_sided("less")
	assert_one_sided("greater")

@mark.parametrize("bad_input",["two-sided","TWO-SIDED","wrzlprmft","LESS","Less"])
def test_assert_one_sided_bad(bad_input):
	with raises(Exception):
		assert_one_sided(bad_input)

def test_zero_p():
	with raises(ValueError):
		CTR( 0, [0,1,2,3] ),

@mark.parametrize(
		"combo",
		chain(*(
			combinations_with_replacement(examples,r)
			for r in range(1,3)
		))
	)
@mark.parametrize("method,variant",combining_statistics)
@mark.parametrize("sampling_method",["proportional","stochastic"])
@mark.parametrize("alt",["less","greater","two-sided"])
def test_commutativity_and_associativity(combo,method,variant,sampling_method,alt,rng):
	get_p = lambda combo,weights: combine(
				combo,
				weights = weights,
				RNG = rng,
				n_samples = n_samples,
				method = method,
				sampling_method = sampling_method,
				alternative = alt,
			).pvalue
	
	n = len(combo)
	combo = np.array(combo)
	weights = rng.random(n) if variant=="weighted" else None
	result_1 = get_p(combo,weights)
	
	new_order = rng.choice(range(n),size=n,replace=False)
	combo = combo[new_order]
	weights = weights[new_order] if variant=="weighted" else None
	result_2 = get_p(combo,weights)
	
	print(combo,weights,result_1,result_2)
	
	assert_matching_p_values(
			result_1,
			result_2,
			n_samples,
			compare = True,
		)

@mark.parametrize("example",examples)
def test_combine_single(example):
	assert combine([example]).pvalue == example.p

# Reproducing a sign test by combining single comparisons:

@mark.parametrize( "n,replicate", product( range(2,15), range(20) ) )
@mark.parametrize( "method", ["fisher","mudholkar_george"] )
def test_comparison_to_sign_test(n,replicate,method,rng):
	def my_sign_test_onesided(X,Y):
		ctrs = [
				CTR( 0.5 if x<y else 1, [0.5,1.0] )
				for x,y in zip(X,Y)
			]
		return combine(
				ctrs,
				n_samples = n_samples,
				method = method,
				RNG = rng,
			).pvalue
	
	X = rng.random(n)
	Y = rng.random(n)
	
	assert_matching_p_values(
		my_sign_test_onesided(X,Y),
		sign_test(X,Y)[0],
		n_samples,
	)

# Reproducing `combine_pvalues` for continuous tests and comparing:

def emulate_continuous_combine_ps(ps,**kwargs):
	ctrs = [ CTR(p) for p in ps ]
	return combine(ctrs,n_samples=n_samples,**kwargs).pvalue

@mark.parametrize( "method", ["fisher","mudholkar_george","stouffer","pearson", "tippett"] )
@mark.parametrize( "n", range(2,17,4) )
@mark.parametrize( "magnitude", ["small","normal"] )
@mark.parametrize("sampling_method",["proportional","stochastic"])
def test_compare_with_combine_pvalues(n,method,magnitude,sampling_method,rng):
	ps = 10**rng.uniform( -3 if magnitude=="small" else -1, 0, n )
	
	assert_matching_p_values(
		emulate_continuous_combine_ps(
			ps,
			method = method,
			RNG = rng,
			sampling_method = sampling_method,
		),
		combine_pvalues(ps,method=method)[1],
		n_samples,
	)

@mark.parametrize( "n", range(2,15) )
@mark.parametrize( "magnitude", ["small","normal"] )
@mark.parametrize("sampling_method",["proportional","stochastic"])
def test_compare_with_combine_pvalues_weighted(n,magnitude,sampling_method,rng):
	ps = 10**rng.uniform( -3 if magnitude=="small" else -1, 0, n )
	weights = rng.random(n)
	
	assert_matching_p_values(
		emulate_continuous_combine_ps(
			ps,
			weights=weights,
			method="stouffer",
			RNG=rng,
			sampling_method = sampling_method,
		),
		combine_pvalues(ps,method="stouffer",weights=weights)[1],
		n_samples,
	)

@mark.parametrize("method,variant",combining_statistics)
@mark.parametrize("variables", ["one", "all"])
@mark.parametrize("sampling_method",["proportional","stochastic"])
@mark.parametrize("alt",["less","greater"])
def test_monotony(method,variant,variables,sampling_method,alt,rng):
	# Test that result increases monotonously with respect to input.
	n,k = 5,7
	changing_values = {
			"less": 10**np.linspace(-3,-0.1,n),
			"greater": 1-10**np.linspace(-3,-0.1,n),
		}[alt]
	pvalues = rng.random(k)
	weights = rng.uniform(0.1,1,size=k)
	combined_ps = []
	errors = []
	for changing_value in changing_values:
		if variables == "one":
			pvalues[0] = changing_value
		else:
			pvalues = np.full(k,changing_value)
		ctrs = [ CTR(p) for p in pvalues ]
		combined_p,error = combine(
					ctrs,
					method = method,
					weights = weights if variant=="weighted" else None,
					alternative = alt,
					n_samples = n_samples,
					sampling_method = sampling_method,
					RNG = rng,
				)
		combined_ps.append(combined_p)
		errors.append(error)
	
	errors = np.array(errors)
	diff_errors = 2*(errors[:-1]+errors[1:])
	
	assert np.all( np.diff(combined_ps) >= -diff_errors )

# CDF of the standard normal distribution and its inverse for Stouffer’s method.
phi = lambda z: (1+erf(z/sqrt(2)))/2
phiinv = lambda x: sqrt(2)*erfinv(2*x-1)

@mark.parametrize(
			"  method  ,   alt      ,  solution   ",
		[
			("tippett" , "less"     , 1-(1-0.4)**3),
			("simes"   , "less"     , 0.9         ),
			("stouffer", "less"     , phi((phiinv(0.4)+phiinv(0.7)+phiinv(0.9))/sqrt(3)) ),
			("tippett" , "greater"  , 1-0.9**3    ),
			("simes"   , "greater"  , 0.3         ),
			("stouffer", "greater"  , phi((phiinv(0.6)+phiinv(0.3)+phiinv(0.1))/sqrt(3)) ),
			("tippett" , "two-sided", (1-0.8**3)  ),
			("stouffer", "two-sided", 2*phi((phiinv(0.6)+phiinv(0.3)+phiinv(0.1))/sqrt(3)) ),
			( lambda p:np.min(p,axis=0), "less" , 1-(1-0.4)**3 ),
		]
	)
@mark.parametrize("sampling_method",["proportional","stochastic"])
def test_simple_case(method,solution,sampling_method,alt,rng):
	assert_matching_p_values(
		emulate_continuous_combine_ps(
			[0.9,0.7,0.4],
			method = method,
			alternative = alt,
			RNG = rng,
			sampling_method = sampling_method,
		),
		solution,
		n_samples
	)

@mark.parametrize("sampling_method",["proportional","stochastic"])
def test_simple_weighted_case(sampling_method,rng):
	assert_matching_p_values(
		emulate_continuous_combine_ps(
			[0.9,0.7,0.4],
			weights = [1,2,3],
			method = "stouffer",
			RNG = rng,
			sampling_method=sampling_method,
		),
		phi( (phiinv(0.9)+2*phiinv(0.7)+3*phiinv(0.4)) / sqrt(1**2+2**2+3**2) ),
		n_samples,
	)

@mark.parametrize("method",(
		method
		for method,variant in combining_statistics
		if variant=="weighted"
	))
@mark.parametrize("sampling_method",["proportional","stochastic"])
@mark.parametrize("alt",["less","greater","two-sided"])
def test_identical_weights(method,sampling_method,alt,rng):
	n = 10
	ps = rng.random(n)
	weights = np.full(n,rng.exponential())
	
	results = [
		emulate_continuous_combine_ps(
			ps,
			RNG=rng,
			method=method,
			weights=w,
			alternative=alt,
			sampling_method=sampling_method,
		)
		for w in [weights,None]
	]
	assert_matching_p_values(*results,n=n_samples,compare=True)

@mark.parametrize("method,variant",combining_statistics)
def test_less_greater_symmetry(method,variant,rng):
	m = 10
	
	ctrs_L,ctrs_G = [],[]
	for _ in range(m):
		n = rng.randint(5,10)
		data = rng.normal(size=n)
		ctrs_L.append( CTR.sign_test( data) )
		ctrs_G.append( CTR.sign_test(-data) )
	
	kwargs = dict(
			weights = rng.random(m) if variant=="weighted" else None,
			method = method,
			n_samples = n_samples,
			RNG = rng,
		)
	result_L = combine(ctrs_L,alternative="less"   ,**kwargs).pvalue
	result_G = combine(ctrs_G,alternative="greater",**kwargs).pvalue
	
	assert_matching_p_values( result_L, result_G, n=n_samples, compare=True )

@mark.parametrize(
		"method_A,method_B",
	[
		( "mudholkar_george", "mudholkar_george" ),
		( "fisher"          , "pearson"          ),
		( "edgington_sym"   , "edgington_sym"    ),
	])
@mark.parametrize("weighted",[False,True])
def test_complement_symmetry(method_A,method_B,weighted,rng):
	m = 10
	
	ctrs_A,ctrs_B = [],[]
	for _ in range(m):
		n = rng.randint(5,10)
		data = rng.normal(size=n)
		ctrs_A.append( CTR.sign_test( data) )
		ctrs_B.append( CTR.sign_test(-data) )
	
	kwargs = dict(
			weights = rng.random(m) if weighted else None,
			n_samples = n_samples,
			RNG = rng,
			alternative = "less",
		)
	result_A =   combine(ctrs_A,method=method_A,**kwargs).pvalue
	result_B = 1-combine(ctrs_B,method=method_B,**kwargs).pvalue
	
	assert_matching_p_values( result_A, result_B, n=n_samples, compare=True )

def test_automatic_weights(rng):
	weights = []
	ctrs = []
	for _ in range(10):
		size_x,size_y = rng.randint(1,10,size=2)
		x = rng.random(size_x)
		y = rng.random(size_y)
		ctrs.append( CTR.mann_whitney_u(x,y,alternative="less") )
		weights.append( size_x + size_y - 1 )
	
	seed = rng.randint(2**32)
	automatic = combine(
			ctrs,
			weights = "dof",
			RNG = np.random.RandomState(seed),
			n_samples = n_samples
		)
	manual = combine(
			ctrs,
			weights = weights,
			RNG = np.random.RandomState(seed),
			n_samples = n_samples
		)
	assert automatic==manual





from .ctr import CTR, combine, direction
from .tools import sign_test

try:
	from .version import version as __version__
except ImportError:
	from warnings import warn
	warn("Failed to find (autogenerated) version.py. Do not worry about this unless you really need to know the version.")



import math
from inspect import signature
import numpy as np
from warnings import warn
from itertools import permutations

from .tools import sign_test, counted_p, Combined_P_Value, is_empty, searchsorted_closest, has_ties
from .pdist import PDist

from scipy.special import erfinv, factorial
from scipy.stats import rankdata, spearmanr, pearsonr, kendalltau, fisher_exact, boschloo_exact, wilcoxon
from scipy.stats._hypotests import _get_wilcoxon_distr
from scipy.stats._mannwhitneyu import _MWU, mannwhitneyu
from scipy.stats._mstats_basic import _kendall_p_exact
from scipy.stats.distributions import hypergeom

def assert_one_sided(alternative):
	if alternative.lower() == "two-sided":
		raise NotImplementedError("The two-sided test is not supported (and makes little sense for combining test results).")
	elif alternative not in ["less","greater"]:
		raise ValueError('Alternative must be "less" or "greater".')

class CTR(object):
	"""
	CTR = combinable test result
	
	Represents a single test result. Use the default constructor to implement a test yourself or use one of the class methods for the respective test.
	
	Parameters
	----------
	p
		The *p* value yielded by the test for the investigated sub-dataset.
	
	all_ps
		An iterable containing all possible *p* values of the test for datasets with the same size as the dataset for this individual test.
		If `None` or empty, all *p* values will be considered possible, i.e., the test will be assumed to be continuous.
	
	dof
		Number of degrees of freedom that affect the test’s outcome; only relevant for automatic weighting (see the `weights` argument of `combine`).
	
	Examples
	--------
	.. code-block:: python3
	
		p = 0.1
		possible_ps = [0.1,0.5,1]
		dof = 7
		ctr = CTR(p,possible_ps,dof)
	
	"""
	def __init__(self,p,all_ps=None,dof=None):
		if p==0: raise ValueError("p value cannot be zero.")
		if np.isnan(p): raise ValueError("p value must not be NaN.")
		
		if not is_empty(all_ps) and p not in all_ps:
			all_ps = np.asarray(all_ps)
			closest = all_ps[np.argmin(np.abs(all_ps-p))]
			if (closest-p)/p > 1e-10:
				raise ValueError(f"p value {p} must be in `all_ps`.")
			else:
				p = closest
		
		self.p = p
		self.nulldist = PDist(all_ps)
		self.q = self.nulldist.complement(self.p)
		self.dof = dof
	
	def __repr__(self):
		return f"""CombinableTestResult(
		\t p-value: {self.p},
		\t nulldist: {self.nulldist}
		\t degrees of freedom: {self.dof}
		\t )"""
	
	def __eq__(self,other):
		return self._approx(other,atol=0)
	
	def _approx(self,other,atol=1e-14):
		"""
		Whether this result is identical to another with in an absolute tolerance `atol` between *p* values.
		"""
		return (
				abs(self.p-other.p) <= atol and
				self.nulldist.approx(other.nulldist,atol) and
				self.dof == other.dof
			)
	
	@classmethod
	def mann_whitney_u( cls, x, y, **kwargs ):
		"""
		Creates an object representing the result of a single Mann–Whitney *U* test (using SciPy’s `mannwhitneyu`).
		
		Ties are not supported yet because I expect them not to occur in the scenarios that require test combinations (but I may be wrong about this) and they make things much more complicated.
		
		For automatic weighting, the number of degrees of freedom is taken to be `len(x) + len(y) - 1`. Note how this aligns with the sign test and Wilcoxon’s signed rank test when either `x` or `y` has size 1.
		
		Parameters
		----------
		x,y
			The two iterables of samples to compare.
		
		kwargs
			Further keyword arguments to be passed on to SciPy’s `mannwhitneyu`, such as `alternative`.
		
		Examples
		--------
		.. code-block:: python3
		
			dataset_A = [20,44,14,68]
			dataset_B = [73,22,80]
			ctr = CTR.mann_whitney_u(dataset_A,dataset_B,alternative="less")
		"""
		x = np.array(x)
		y = np.array(y)
		
		if "alternative" not in kwargs:
			raise ValueError("You must specify the alternative.")
		assert_one_sided(kwargs["alternative"])
		
		if np.any(x[:,None]==y):
			raise NotImplementedError("Ties are not yet implemented.")
		
		n,m = len(x),len(y)
		
		if kwargs.pop("method","exact") != "exact":
			warn('Can only use `method="exact"`.')
		
		p = mannwhitneyu(x,y,method="exact",**kwargs).pvalue
		possible_ps = np.cumsum( _MWU(n,m).build_u_freqs_array(n*m) )
		return cls( p, possible_ps, n+m-1 )
	
	@classmethod
	def sign_test( cls, x, y=0, alternative="less" ):
		"""
		Creates an object representing the result of a single sign test.
		
		For automatic weighting, the number of pairs is taken as the number of degrees of freedom.
		
		Parameters
		----------
		x,y
			The two arrays of paired samples to compare. If `y` is a number, a one-sample sign test is performed with `y` as the median. With `y` as an iterable, the two-sample test is performed.
		
		alternative: "less" or "greater"
		
		Examples
		--------
		.. code-block:: python3
		
			dataset_A = [20,44,14,68]
			dataset_B = [73,22,80,53]
			ctr = CTR.sign_test(dataset_A,dataset_B,alternative="less")
		"""
		
		assert_one_sided(alternative)
		
		p,m,_ = sign_test(x,y,alternative)
		
		all_ps = list( np.cumsum([math.comb(m,i)/2**m for i in range(m)]) ) + [1]
		
		return cls( p, all_ps, m )
	
	@classmethod
	def wilcoxon_signed_rank( cls, x, y=None, alternative="less" ):
		"""
		Creates an object representing the result of a single Wilcoxon signed-rank test.
		
		For automatic weighting, the number of pairs is taken as the number of degrees of freedom.
		
		Parameters
		----------
		x,y
			The two arrays of paired samples to compare. If `y` is `None`, the one-sample test is performed, otherwise the two-sample one.
		
		alternative: "less" or "greater"
		
		Examples
		--------
		.. code-block:: python3
		
			dataset_A = [20,44,14,68]
			dataset_B = [73,22,80,53]
			ctr = CTR.wilcoxon_signed_rank(dataset_A,dataset_B,alternative="less")
		"""
		
		assert_one_sided(alternative)
		
		d = np.asarray(x) - (y or 0)
		if has_ties(np.abs(d)) or np.any(d==0):
			raise NotImplementedError("Ties and zeros are not yet implemented.")
		
		n = len(x)
		all_ps = np.cumsum( _get_wilcoxon_distr(n) )
		p = wilcoxon(x,y,alternative=alternative,mode="exact").pvalue
		
		return cls( p, all_ps, n )
	
	@classmethod
	def spearmanr( cls, x, y, alternative="greater", n_thresh=9 ):
		"""
		Creates an object representing the result of a single Spearman’s ρ test.
		If the size of `x` and `y`, *n,* is smaller than `n_thresh`, *p* values are exactly determined using a permutation test. Otherwise *p* values are computed using SciPy’s `spearmanr` assuming a uniform distribution of *p* values and ensuring :math:`p≥\\frac{1}{n!}`.
		
		For automatic weighting, the number of degrees of freedom is taken to be one less than the number of pairs.
		
		Parameters
		----------
		x,y
			The two arrays of samples to correlate.
		
		alternative: "greater" or "less"
		
		n_thresh:
			Threshold under which a permutation test is used.
		
		Examples
		--------
		.. code-block:: python3
		
			dataset_A = [1,3,4,2,5,6]
			dataset_B = [3,1,2,5,6,4]
			ctr = CTR.spearmanr(dataset_A,dataset_B,alternative="greater")
		"""
		n = len(x)
		
		if n>n_thresh:
			p = spearmanr(x,y,alternative=alternative).pvalue
			p = np.clip( p, 1/factorial(n), 1 )
			return cls(p,dof=n-1)
		
		# Working with n³·cov(2R(x),2R(y)) because it is integer. As a statistics, it is equivalent to Spearman’s ρ.
		x_r = np.fix(2*rankdata(x)).astype(int)
		y_r = np.fix(2*rankdata(y)).astype(int)
		x_normed = n*x_r - np.sum(x_r)
		y_normed = n*y_r - np.sum(y_r)
		
		orig_cov = np.sum(x_normed*y_normed)
		possible_covs = np.sort([
				np.sum(x_normed*y_permut)
				for y_permut in permutations(y_normed)
			])
		
		if alternative == "greater":
			possible_covs = np.flip(possible_covs)
		assert_one_sided(alternative)
		
		k = len(possible_covs)
		# Using the last of duplicate covs by updating dictionary in the right order:
		cov_to_p = dict( zip( possible_covs, np.linspace(1/k,1,k) ) )
		
		orig_p = cov_to_p[orig_cov]
		return cls( orig_p, list(cov_to_p.values()), n-1 )
	
	@classmethod
	def kendalltau( cls, x, y, **kwargs ):
		"""
		Creates an object representing the result of a single Kendall’s τ test using SciPy’s `kendalltau` to compute *p* values.
		
		NaNs and ties are not supported yet.
		
		For automatic weighting, the number of degrees of freedom is taken to be one less than the number of pairs.
		
		Parameters
		----------
		x,y
			The two arrays of samples to correlate.
		
		alternative: "greater" or "less"
		
		Examples
		--------
		.. code-block:: python3
		
			dataset_A = [1,3,4,2,5,6]
			dataset_B = [3,1,2,5,6,4]
			ctr = CTR.kendalltau(dataset_A,dataset_B,alternative="greater")
		"""
		
		assert_one_sided(kwargs["alternative"])
		
		if has_ties(x) or has_ties(y):
			raise NotImplementedError("Ties are not yet implemented.")
		
		p = kendalltau(x,y,**kwargs).pvalue
		n = len(x)
		tot = math.comb(n,2)
		
		possible_ps = [
			_kendall_p_exact(n,dis,"greater")
			for dis in range(0,math.comb(n,2)+1)
		]
		
		return cls(p,possible_ps,n-1)
	
	@classmethod
	def fisher_exact( cls, C, alternative="less" ):
		"""
		Creates an object representing the result of Fisher’s exact test for a single contingency table `C`. This is unrelated to Fisher’s method of combining *p* values. Note that, for most scientific applications, the restrictive conditions of this test are not met and Boschloo’s exact test is more appropriate.
		
		For automatic weighting, the sum over `C` is taken as the number of degrees of freedom.
		
		Parameters
		----------
		C: 2×2 array or nested iterable
			The contingency table.
		
		alternative: "less" or "greater"
		
		Examples
		--------
		.. code-block:: python3
		
			contingency_table = [[10,4],[5,4]]
			ctr = CTR.fisher_exact(contingency_table,alternative="greater")
		"""
		
		assert_one_sided(alternative)
		C = np.fliplr(C) if alternative=="greater" else np.array(C)
		
		p = fisher_exact(C,alternative="less")[1]
		
		n1,n2 = np.sum(C,axis=1)
		n ,_  = np.sum(C,axis=0)
		
		possible_ps = [
				hypergeom.cdf( x, n1+n2, n1, n )
				for x in range( max(0,n-n2), min(n,n1)+1 )
			]
		
		return cls( p, possible_ps, np.sum(C) )
	
	@classmethod
	def boschloo_exact( cls, C, alternative="less", n=32, atol=1e-10 ):
		"""
		Creates an object representing the result of Boschloo’s exact for a single contingency table `C` using SciPy’s implementation.
		
		For automatic weighting, the sum over `C` is taken as the number of degrees of freedom.
		
		Parameters
		----------
		C: 2×2 array or nested iterable
			The contingency table.
		
		alternative: "less" or "greater"
		
		n
			The same parameter of SciPy’s `boschloo_exact`.
		
		atol
			*p* values that are closer than this are treated as identical.
		
		Examples
		--------
		.. code-block:: python3
		
			contingency_table = [[10,4],[5,4]]
			ctr = CTR.boschloo_exact(contingency_table,alternative="greater")
		"""
		
		assert_one_sided(alternative)
		C = np.fliplr(C) if alternative=="greater" else np.array(C)
		
		p = boschloo_exact(C,alternative="less",n=n).pvalue
		
		n1,n2 = np.sum(C,axis=1)
		
		possible_ps = sorted(
				boschloo_exact(
						[ [ C11, n1-C11 ], [ C21, n2-C21 ] ],
						alternative="less",
						n=n,
					).pvalue
				for C11 in range( 0, n1+1 )
				for C21 in range( C11==0, n2+(C11!=n1) )
			)
		
		# Unify close p values.
		i = 1
		while i<len(possible_ps):
			if possible_ps[i-1]+atol > possible_ps[i]:
				del possible_ps[i]
			else:
				i += 1
		
		return cls( p, possible_ps, np.sum(C) )

combining_statistics = {
	("fisher"          ,"normal"  ): lambda p:  np.sum( np.log(p)     , axis=0 ),
	("pearson"         ,"normal"  ): lambda q: -np.sum( np.log(q)     , axis=0 ),
	("mudholkar_george","normal"  ): lambda p,q:np.sum( np.log(p/q)   , axis=0 ),
	("stouffer"        ,"normal"  ): lambda p:  np.sum( erfinv(2*p-1) , axis=0 ),
	("tippett"         ,"normal"  ): lambda p:  np.min( p             , axis=0 ),
	("edgington"       ,"normal"  ): lambda p:  np.sum( p             , axis=0 ),
	("edgington_sym"   ,"normal"  ): lambda p,q:np.sum( p-q           , axis=0 ),
	("simes"           ,"normal"  ): lambda p:  np.min(p/rankdata(p,axis=0,method="ordinal"),axis=0),
	("fisher"          ,"weighted"): lambda p,w:    w.dot(np.log(p))     ,
	("pearson"         ,"weighted"): lambda q,w:   -w.dot(np.log(q))     ,
	("mudholkar_george","weighted"): lambda p,q,w:  w.dot(np.log(p/q))   ,
	("stouffer"        ,"weighted"): lambda p,w:    w.dot(erfinv(2*p-1)) ,
	("edgington"       ,"weighted"): lambda p,w:    w.dot(p)             ,
	("edgington_sym"   ,"weighted"): lambda p,q,w:  w.dot(p+1-q)         ,
}

statistics_with_inf = {"stouffer"}

def flip_pq(args):
	if isinstance(args,str) and len(args)==1:
		if args == "p":
			return "q"
		elif args == "q":
			return "p"
		else:
			return args
	else:
		return { flip_pq(arg) for arg in args }

def apply_statistics(statistic,data,alternative="less"):
	if alternative in ["less","greater"]:
		kwargs = {
			par: data[ par if alternative=="less" else flip_pq(par) ]
			for par in signature(statistic).parameters
		}
		return statistic(**kwargs)
	elif alternative == "two-sided":
		return np.minimum(
				apply_statistics(statistic,data,"less"   ),
				apply_statistics(statistic,data,"greater"),
			)
	else:
		raise ValueError('Alternative must be "less", "greater", or "two-sided".')

def get_statistic(method,weights):
	if method in (method for method,_ in combining_statistics):
		if weights is None:
			return combining_statistics[method,"normal"]
		else:
			try:
				return combining_statistics[method,"weighted"]
			except KeyError:
				raise ValueError(f'No weighted version of "{method}" method')
	else:
		if not callable(method):
			raise ValueError(f'Method "{method}" is neither known nor callable.')
		return method

def combine(
		ctrs, weights=None,
		method="mudholkar_george", alternative="less",
		n_samples=10000000, sampling_method="proportional",
		rtol=1e-15, atol=1e-15,
		RNG=None,
	):
	"""
	Estimates the combined *p* value of combinable test results. Usually, this result is why you are using this module.
	
	Parameters
	----------
	ctrs: iterable of CTRs
		The test results that shall be combined.
	
	method: string or function
		One of "fisher", "pearson", "mudholkar_george", "stouffer", "tippett", "edgington", "edgington_sym", "simes", or a self-defined function.
		
		In the latter case, the function can have the following arguments (which must be named as given):
		
		* A two-dimensional array `p` containing the *p* values.
		* A two-dimensional array `q` containing their complements.
		* A one-dimensional array `w` containing the weights.
		
		The function must return the statistics computed along the zero-th axis.
		For example for the weighted Mudholkar–George method, this function would be `lambda p,q,w:  w.dot(np.log(p/q))`.
		The sign of the statistics must be such that low values indicate a high significance.
	
	alternative: "less", "greater", or "two-sided"
		The direction of the (common) trend that your compound null hypothesis is testing against.
		Mind that this is not about the sidedness of the individual tests: Those should always be one-sided.
		
		* If "less", the compound research hypothesis is that the subtests exhibit a trend towards a low *p* value.
		* If "greater", the compound research hypothesis is that the subtests exhibit a trend towards high *p* values (close to 1). In this case, the method of choice will be applied to the complements of the *p* values (see `complements`).
		* If "two-sided", the compound research hypothesis is that the subtests exhibit either of the two above trends. Beware that this is not necessarily the same as just doubling the *p* value of the respective one-sided test, since for some combining methods, a compound dataset may exhibit **both** trends.
	
	weights: iterable of numbers or "dof"
		Weights for individual results. Does not work for minimum-based methods (Tippett and Simes).
		
		If `"dof"`, each test will be weighted with its degrees of freedom, i.e., the number of samples that can independently affect the test’s result. As there is some potential for ambiguity as to what the degrees of freedom of a given test are, please check that they are what you expect by looking at the tests’ documentations or the attribute `dof` of a CombinableTestResult. This particularly applies when combining different tests.
	
	n_samples
		Number of samples used for Monte Carlo simulation. High numbers increase the accuracy, but also the runtime and memory requirements.
	
	rtol: non-negative float
	atol: non-negative float
		Values of the statistics closer than specified by `atol` and `rtol` are regarded as identical (as in `numpy.isclose`). A small value (such as the default) may improve the results if numerical noise makes values different.
	
	RNG
		NumPy random-number generator used for the Monte Carlo simulation.
		If `None`, it will be automatically generated.
	
	sampling_method: "proportional" or "stochastic"
		How to sample from the *p* value distributions of the subtests.
		
		If `"proportional"`, the frequency of *p* values for each individual result will be exactly proportional to its probability – except for rounding. Only the rounding and the order of elements will be random.
		
		If `"stochastic"`, the values will be randomly sampled and thus their sampled frequencies are subject to stochastic fluctuations. This usually leads to slightly less accurate results, but the simulations are statistically independent.
		
		The author of these lines cannot think of any disadvantage to the first approach and has not found any in numerical experiments.
	
	Returns
	-------
	pvalue
		The estimated combined *p* value.
	
	std
		The estimated standard deviation of *p* values when repeating the sampling. This is accurate for stochastic sampling and overestimating for proportional sampling.
	
	Examples
	--------
	.. code-block:: python3
	
		ctrs = [
			CTR.sign_test( [24,58,10], [65,51,61], alternative="less" ),
			CTR.mann_whitney_u( [20,44,14,68], [73,22,80], alternative="less" ),
			CTR( ttest_ind( [28,36,11], [93,76,70,83], alternative="less" ).pvalue ),
		]
		compound_p = combine(ctrs,alternative="less").pvalue
	"""
	
	if len(ctrs)==1:
		p = ctrs[0].p
		q = ctrs[0].q
		if alternative=="less":
			return Combined_P_Value(p,0)
		elif alternative=="greater":
			return Combined_P_Value(q,0)
		elif alternative=="two-sided":
			return Combined_P_Value( min(2*p,2*q,1), 0 )
	
	statistic = get_statistic(method,weights)
	
	# required_args is a subset of {"p","q","w"} indicating the arguments of the combining statistic.
	required_args = set(signature(statistic).parameters)
	if alternative == "greater":
		required_args = flip_pq(required_args)
	elif alternative == "two-sided":
		required_args = required_args | flip_pq(required_args)
	
	sampling_kwargs = dict(RNG=RNG,size=n_samples,method=sampling_method)
	
	data_null = {}
	if {"p","q"} <= required_args:
		data_null["p"] = np.empty((len(ctrs),n_samples))
		data_null["q"] = np.empty((len(ctrs),n_samples))
		for ctr,target_p,target_q in zip(ctrs,data_null["p"],data_null["q"]):
			# target[:] to overwrite the content of target instead of reassigning the variable.
			target_p[:],target_q[:] = ctr.nulldist.sample_both(**sampling_kwargs)
	else:
		for x in {"p","q"} & required_args:
			data_null[x] = np.empty((len(ctrs),n_samples))
			for ctr,target in zip(ctrs,data_null[x]):
				target[:] = ctr.nulldist.sample(which=x,**sampling_kwargs)
	
	data_orig = {
			x : np.array([getattr(ctr,x) for ctr in ctrs])
			for x in ["p","q"]
		}
	
	if isinstance(weights,str) and weights == "dof":
		weights = [ ctr.dof for ctr in ctrs ]
	if weights is not None:
		data_null["w"] = data_orig["w"] = np.asarray(weights)
	
	err_kwargs = {"divide":"ignore","invalid":"ignore"} if (method in statistics_with_inf) else {}
	with np.errstate(**err_kwargs):
		orig_stat  = apply_statistics(statistic,data_orig,alternative=alternative)
		null_stats = apply_statistics(statistic,data_null,alternative=alternative)
	
	return counted_p( orig_stat, null_stats, rtol=rtol, atol=atol )

def direction( ctrs, weights=None, method="mudholkar_george" ):
	"""
	A service function to indicate whether the `ctrs` are rather trending towards high or low *p* values.
	
	If you are combining two-sidedly, this tells you the direction of the strongest trend, whether it’s significant or not (that’s what `combine` is for). Beware that for some methods such as Fisher’s, a compound dataset may exhibit a significant trend in **both** directions and this function won’t tell you. This cannot happen for symmetric methods (Mudholkar–George, Stouffer, and symmetrised Edgington).
	
	Parameters
	----------
	`weights` and `method` as for `combine`.
	
	Returns
	-------
	direction: string
		One of "less", "greater" or "equal". The latter is only returned if the statistics of the combining method are exactly equal in either direction.
	
	Examples
	--------
	.. code-block:: python3
	
		ctrs = [
			CTR.sign_test( [24,58,10], [65,51,61], alternative="less" ),
			CTR.mann_whitney_u( [20,44,14,68], [73,22,80], alternative="less" ),
			CTR( ttest_ind( [28,36,11], [93,76,70,83], alternative="less" ).pvalue ),
		]
		trend = direction(ctrs)
	"""
	
	if len(ctrs)==1:
		stats = { "less": ctrs[0].p, "greater": ctrs[0].q }
	else:
		statistic = get_statistic(method,weights)
		
		data_orig = {
				x : np.array([getattr(ctr,x) for ctr in ctrs])
				for x in ["p","q"]
			}
		
		if isinstance(weights,str) and weights == "dof":
			weights = [ ctr.dof for ctr in ctrs ]
		if weights is not None:
			data_orig["w"] = np.asarray(weights)
		
		err_kwargs = {"divide":"ignore","invalid":"ignore"} if (method in statistics_with_inf) else {}
		with np.errstate(**err_kwargs):
			stats = {
				case: apply_statistics(statistic,data_orig,alternative=case)
				for case in ["less","greater"]
			}
	
	if stats["less"] < stats["greater"]:
		return "less"
	elif stats["greater"] < stats["less"]:
		return "greater"
	else:
		return "equal"




import numpy as np
from .tools import is_empty

def sample_discrete(values,frequencies,RNG=None,size=10000000,method="proportional"):
	"""
	Returns `size` samples from `values` with `frequencies` using `RNG` as the random-number generator.
	
	If `method` is `"proportional"`, the frequency of each value will be exact – except for rounding. Only the rounding and the order of elements will be stochastic.
	
	If `method` is `"stochastic"`, the values will be randomly sampled and thus their actual frequencies are subject to stochastic fluctuations. This usually leads to slightly less accurate results, but independent samples.
	"""
	RNG = RNG or np.random.default_rng()
	
	if method=="stochastic":
		return RNG.choice( values, p=frequencies, size=size, replace=True )
	elif method=="proportional":
		combos = list(zip(values,frequencies))
		# Shuffling to randomise the effect of rounding errors:
		RNG.shuffle(combos)
		result = np.empty(size)
		start = 0
		for p,prob in combos:
			end = start + prob*size
			result[ round(start) : round(end) ] = p
			start = end
		assert round(end) == size
		RNG.shuffle(result)
		return result
	else:
		raise ValueError('Method must either be "proportional" or "stochastic"')

def sample_uniform(RNG=None,size=10000000,method="proportional"):
	"""
	Returns `size` samples from the continuous uniform distribution on [0,1) using `RNG` as the random-number generator.
	
	If `method` is `"proportional"`, the values will be evenly spaced (but random in order).
	
	If `method` is `"stochastic"`, the values will be randomly sampled, but independent.
	"""
	
	RNG = RNG or np.random.default_rng()
	
	if method=="stochastic":
		return RNG.uniform(size=size)
	elif method=="proportional":
		pad = 1/(2*size)
		result = np.linspace( pad, 1-pad, size )
		RNG.shuffle(result)
		return result
	else:
		raise ValueError('Method must either be "proportional" or "stochastic"')

class PDist(object):
	"""
	Represents a uniform distribution (of p values) on the unit interval with a specified support, i.e., a distribution with :math:`\\mathop{CDF}(p)=p` for any :math:`p` in the support.
	For any test, the p values follow such a distribution under the null hypothesis.
	All you need to know are the possible p values.
	
	Parameters
	----------
	ps
		iterable containing the p values that are the support of the new distribution.
		If empty, this represents the continous uniform distribution.
	"""
	def __init__(self,ps):
		self.ps = np.array([]) if is_empty(ps) else np.atleast_1d(ps)
		self.ps.sort()
		if not self.continuous:
			if not ( ( 0 < self.ps[0] ) and ( abs(self.ps[-1]-1) < 1e-10 ) ):
				raise ValueError(f"p values must be between 0 and 1, with the largest being 1; but they are {ps}")
			if len(self.ps)>1 and self.ps[-2]>=1:
				raise ValueError("Two p values slightly larger than or equal to 1.")
			
			# Get rid of any numerical inaccuracies:
			self.ps[-1] = 1
	
	@property
	def probs(self):
		return np.diff(self.ps,prepend=0)
	
	@property
	def continuous(self):
		return self.ps.size == 0
	
	def __iter__(self):
		yield from self.ps
	
	def __repr__(self):
		if self.continuous:
			return f"PDist( uniform )"
		else:
			points = ", ".join(f"{p:.3g}" for p in self)
			return f"PDist( {points} )"
	
	def __eq__(self,other):
		return self.approx(other,atol=0)
	
	def approx(self,other,atol=1e-14):
		"""
		Whether this distribution is identical to another with in an absolute tolerance `atol` between *p* values.
		"""
		if self.continuous:
			return other.continuous
		else:
			return all( abs(p1-p2)<=atol for p1,p2 in zip(self,other) )
	
	def sample(self,which="p",**kwargs):
		"""
		Returns `size` samples from the p or q values (specified by `which`), using `RNG` as the random-number generator. See sample_discrete and sample_uniform for arguments.
		"""
		if which=="p":
			return self.sample_ps(**kwargs)
		elif which=="q":
			return self.sample_complement(**kwargs)
	
	def sample_ps(self,which="p",**kwargs):
		"""
		Returns `size` samples from the distribution using `RNG` as the random-number generator. See sample_discrete and sample_uniform for arguments.
		"""
		if self.continuous:
			return 1-sample_uniform(**kwargs)
		else:
			return sample_discrete(self.ps,self.probs,**kwargs)
	
	@property
	def complement_ps(self):
		comp_ps = 1-np.roll(self.ps,1)
		comp_ps[0] = 1
		return comp_ps

	def sample_complement(self,**kwargs):
		"""
		Returns `size` samples from the complementary values of the distribution using `RNG` as the random-number generator. See sample_discrete and sample_uniform for arguments.
		"""
		if self.continuous:
			return sample_uniform(**kwargs)
		else:
			return sample_discrete(
					self.complement_ps,
					self.probs,
					**kwargs,
				)
	
	def sample_both(self,**kwargs):
		"""
		Returns `size` samples from the values and the complementary values of the distribution using `RNG` as the random-number generator. These are paired. See sample_discrete and sample_uniform for arguments.
		"""
		if self.continuous:
			qs = sample_uniform(**kwargs)
			return 1-qs, qs
		else:
			indices = sample_discrete(
					np.arange(len(self.ps)),
					self.probs,
					**kwargs,
				).astype(int)
			return self.ps[indices], self.complement_ps[indices]
	
	def complement(self,pvalue):
		"""
		Returns the complement of a particular p value with respect to the distribution, i.e., the probability that a value is larger or equal than the given value.
		"""
		if self.continuous:
			return 1-pvalue
		else:
			pos = np.argmin(np.abs(self.ps-pvalue))
			if pos!=0:
				return 1-self.ps[pos-1]
			else:
				return 1




#!/usr/bin/python3
# -*- coding: utf-8 -*-

"""
We have three pairs of independent datasets `(A_1,A_2)`, `(B_1,B_2)`, and `(C_1,C_2)`.
Our research hypothesis is that the values in the respective first dataset are lower.
However, the datasets have different properties and thus cannot simply be pooled:

* `(A_1,A_2)` is a paired dataset. Since we cannot assume any distribution, we want to apply the sign test.
* `(B_1,B_2)` is an unpaired dataset (with unequal sizes). Again, we cannot assume a distribution. Hence we want to apply the Mann–Whitney *U* test.
* `(C_1,C_2)` is a paired dataset, however we can assume that the differences to be normally distributed and thus apply the *t* test for two dependent samples.

We start with performing the sign test on `A_1` and `A_2`.
We create a `CTR` from this:

.. literalinclude:: ../examples/simple_example.py
	:start-after: example-st\u0061rt
	:dedent: 1
	:lines: 1-2

`result_A` now contains the *p* value of the sign test as well as the null distribution of possible *p* values.
It is ready for being combined with other test results, but we have to create these first.
We do so by doing something very similar with `B_1` and `B_2` and the Mann–Whitney *U* test:

.. literalinclude:: ../examples/simple_example.py
	:start-after: example-st\u0061rt
	:dedent: 1
	:lines: 4-4

Finally, we perform the *t* test on `C_1` and `C_2`.
Since the *t* test is a continuous test, we do not need a special constructor to create a `CTR`, but can use generic one using only the *p* value computed with an existing function, here `scipy.stats.ttest_rel`.
Leaving the second argument empty specifies that the test is continuous:

.. literalinclude:: ../examples/simple_example.py
	:start-after: example-st\u0061rt
	:dedent: 1
	:lines: 6-8

After we have performed the individual tests, we can obtain the compound *p* value using `combine` (which we imported earlier).

.. literalinclude:: ../examples/simple_example.py
	:start-after: example-st\u0061rt
	:dedent: 1
	:lines: 10-10
"""

if __name__ == "__main__":
	import numpy as np
	
	A_1,A_2 = np.random.random((2,10))
	B_1 = np.random.random(7)
	B_2 = np.random.random(8)
	C_1,C_2 = np.random.normal(size=(2,10))
	
	# example-start
	from combine_pvalues_discrete import CTR, combine
	result_A = CTR.sign_test(A_1,A_2,alternative="less")
	
	result_B = CTR.mann_whitney_u(B_1,B_2,alternative="less")
	
	from scipy.stats import ttest_rel
	p_C = ttest_rel(C_1,C_2).pvalue
	result_C = CTR(p_C)
	
	print( combine([result_A,result_B,result_C]) )




#!/usr/bin/python3

from setuptools import setup
from io import open

setup(
		name = "combine_pvalues_discrete",
		description = "A Python toolbox for combining p values from tests with a discrete statistics",
		long_description = open("README.rst", encoding="utf8").read(),
		python_requires = ">=3.7",
		packages = ["combine_pvalues_discrete"],
		install_requires = ["numpy","scipy>=1.14"],
		setup_requires = ["setuptools_scm","pytest-runner","pytest-rng"],
		extras_require = {'test': ['pytest','statsmodels','pytest-rng']},
		use_scm_version = {"write_to": "combine_pvalues_discrete/version.py", "local_scheme": "no-local-version"},
		classifiers = [
				"License :: OSI Approved :: BSD License",
				"Operating System :: POSIX",
				"Operating System :: MacOS :: MacOS X",
				"Operating System :: Microsoft :: Windows",
				"Programming Language :: Python",
				"Topic :: Scientific/Engineering",
			],
	)

